"use strict";
/**
 * AiScript interpreter
 */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
const autobind_decorator_1 = require("autobind-decorator");
const scope_1 = require("./scope");
const error_1 = require("./error");
const std_1 = require("./lib/std");
const util_1 = require("./util");
const value_1 = require("./value");
class AiScript {
    constructor(vars, opts) {
        this.stepCount = 0;
        this.stop = false;
        this.abortHandlers = [];
        this.opts = opts || {};
        const io = {
            print: value_1.FN_NATIVE(args => {
                if (this.opts.out)
                    this.opts.out(args[0]);
            }),
            readline: value_1.FN_NATIVE(async (args) => {
                const q = args[0];
                util_1.assertString(q);
                if (this.opts.in == null)
                    return value_1.NULL;
                const a = await this.opts.in(q.value);
                return value_1.STR(a);
            })
        };
        this.vars = Object.assign(Object.assign(Object.assign({}, vars), std_1.std), io);
        this.scope = new scope_1.Scope([new Map(Object.entries(this.vars))]);
        this.scope.opts.log = (type, params) => {
            switch (type) {
                case 'add':
                    this.log('var:add', params);
                    break;
                case 'read':
                    this.log('var:read', params);
                    break;
                case 'write':
                    this.log('var:write', params);
                    break;
                default: break;
            }
        };
    }
    async exec(script) {
        if (script == null || script.length === 0)
            return;
        await this.collectNs(script);
        const result = await this._run(script, this.scope);
        this.log('end', { val: result });
    }
    async execFn(fn, args) {
        return this._fn(fn, args);
    }
    static collectMetadata(script) {
        if (script == null || script.length === 0)
            return;
        function nodeToJs(node) {
            switch (node.type) {
                case 'arr': return node.value.map(item => nodeToJs(item));
                case 'bool': return node.value;
                case 'null': return null;
                case 'num': return node.value;
                case 'obj': {
                    const obj = {};
                    for (const [k, v] of node.value.entries()) {
                        // TODO: keyが__proto__とかじゃないかチェック
                        obj[k] = nodeToJs(v);
                    }
                    return obj;
                }
                case 'str': return node.value;
            }
        }
        const meta = new Map();
        for (const node of script) {
            switch (node.type) {
                case 'meta': {
                    meta.set(node.name, nodeToJs(node.value));
                    break;
                }
                default: {
                    // nop
                }
            }
        }
        return meta;
    }
    log(type, params) {
        if (this.opts.log)
            this.opts.log(type, params);
    }
    async collectNs(script) {
        for (const node of script) {
            switch (node.type) {
                case 'ns': {
                    await this.collectNsMember(node);
                    break;
                }
                default: {
                    // nop
                }
            }
        }
    }
    async collectNsMember(ns) {
        const scope = this.scope.createChildScope();
        for (const node of ns.members) {
            switch (node.type) {
                case 'def': {
                    const v = await this._eval(node.expr, scope);
                    scope.add(node.name, v);
                    this.scope.add(ns.name + ':' + node.name, v);
                    break;
                }
                case 'ns': {
                    break; // TODO
                }
                default: {
                    throw new Error('invalid ns member type: ' + node.type);
                }
            }
        }
    }
    async _fn(fn, args) {
        if (fn.native) {
            const result = await Promise.resolve(fn.native(args, {
                call: this._fn,
                registerAbortHandler: this.registerAbortHandler,
                unregisterAbortHandler: this.unregisterAbortHandler,
            }));
            return result || value_1.NULL;
        }
        else {
            const _args = new Map();
            for (let i = 0; i < (fn.args || []).length; i++) {
                _args.set(fn.args[i], args[i]);
            }
            const fnScope = fn.scope.createChildScope(_args);
            return value_1.unWrapRet(await this._run(fn.statements, fnScope));
        }
    }
    async _eval(node, scope) {
        if (this.stop)
            return value_1.NULL;
        this.log('node', { node: node });
        this.stepCount++;
        if (this.opts.maxStep && this.stepCount > this.opts.maxStep) {
            throw new error_1.AiScriptError('max step exceeded');
        }
        switch (node.type) {
            case 'call': {
                const val = scope.get(node.name);
                util_1.assertFunction(val);
                const args = await Promise.all(node.args.map(async (expr) => await this._eval(expr, scope)));
                return this._fn(val, args);
            }
            case 'if': {
                const cond = await this._eval(node.cond, scope);
                util_1.assertBoolean(cond);
                if (cond.value) {
                    return this._eval(node.then, scope);
                }
                else {
                    if (node.elseif && node.elseif.length > 0) {
                        for (const elseif of node.elseif) {
                            const cond = await this._eval(elseif.cond, scope);
                            util_1.assertBoolean(cond);
                            if (cond.value) {
                                return this._eval(elseif.then, scope);
                            }
                        }
                        if (node.else) {
                            return this._eval(node.else, scope);
                        }
                    }
                    else if (node.else) {
                        return this._eval(node.else, scope);
                    }
                }
                return value_1.NULL;
            }
            case 'match': {
                const about = await this._eval(node.about, scope);
                for (const qa of node.qs) {
                    const q = await this._eval(qa.q, scope);
                    if (util_1.eq(about, q)) {
                        return await this._eval(qa.a, scope);
                    }
                }
                if (node.default) {
                    return await this._eval(node.default, scope);
                }
                return value_1.NULL;
            }
            case 'for': {
                let result = [];
                if (node.times) {
                    const times = await this._eval(node.times, scope);
                    util_1.assertNumber(times);
                    for (let i = 1; i < times.value + 1; i++) {
                        const v = await this._eval(node.for, scope);
                        result.push(v);
                    }
                }
                else {
                    const from = await this._eval(node.from, scope);
                    const to = await this._eval(node.to, scope);
                    util_1.assertNumber(from);
                    util_1.assertNumber(to);
                    for (let i = from.value + 1; i < to.value + 1; i++) {
                        const v = await this._eval(node.for, scope.createChildScope(new Map([
                            [node.var, value_1.NUM(i)]
                        ])));
                        result.push(v);
                    }
                }
                return value_1.ARR(result);
            }
            case 'forOf': {
                let result = [];
                const items = await this._eval(node.items, scope);
                util_1.assertArray(items);
                for (const item of items.value) {
                    const v = await this._eval(node.for, scope.createChildScope(new Map([
                        [node.var, item]
                    ])));
                    result.push(v);
                }
                return value_1.ARR(result);
            }
            case 'def': {
                scope.add(node.name, await this._eval(node.expr, scope));
                return value_1.NULL;
            }
            case 'var': {
                return scope.get(node.name);
            }
            case 'assign': {
                scope.assign(node.name, await this._eval(node.expr, scope));
                return value_1.NULL;
            }
            case 'propAssign': {
                const obj = scope.get(node.obj);
                let x = obj;
                const lastProp = node.path[node.path.length - 1];
                for (const prop of node.path.slice(0, node.path.length - 1)) {
                    util_1.assertObject(x);
                    if (!x.value.has(prop)) {
                        x = value_1.NULL;
                        break;
                    }
                    else {
                        x = x.value.get(prop);
                    }
                }
                util_1.assertObject(x);
                x.value.set(lastProp, await this._eval(node.expr, scope));
                return value_1.NULL;
            }
            case 'indexAssign': {
                const arr = scope.get(node.arr);
                util_1.assertArray(arr);
                const i = await this._eval(node.i, scope);
                util_1.assertNumber(i);
                arr.value[i.value - 1] = await this._eval(node.expr, scope); // TODO: 存在チェック
                return value_1.NULL;
            }
            case 'inc': {
                const target = scope.get(node.name);
                util_1.assertNumber(target);
                const v = await this._eval(node.expr, scope);
                util_1.assertNumber(v);
                target.value += v.value;
                return value_1.NULL;
            }
            case 'dec': {
                const target = scope.get(node.name);
                util_1.assertNumber(target);
                const v = await this._eval(node.expr, scope);
                util_1.assertNumber(v);
                target.value -= v.value;
                return value_1.NULL;
            }
            case 'null': return value_1.NULL;
            case 'bool': return value_1.BOOL(node.value);
            case 'num': return value_1.NUM(node.value);
            case 'str': return value_1.STR(node.value);
            case 'arr': return value_1.ARR(await Promise.all(node.value.map(async (item) => await this._eval(item, scope))));
            case 'obj': {
                const obj = new Map();
                for (const k of node.value.keys()) {
                    obj.set(k, await this._eval(node.value.get(k), scope));
                }
                return value_1.OBJ(obj);
            }
            case 'prop': {
                const obj = scope.get(node.obj);
                let x = obj;
                for (const prop of node.path) {
                    util_1.assertObject(x);
                    if (!x.value.has(prop)) {
                        x = value_1.NULL;
                        break;
                    }
                    else {
                        x = x.value.get(prop);
                    }
                }
                return x;
            }
            case 'propCall': {
                const obj = scope.get(node.obj);
                let x = obj;
                for (const prop of node.path) {
                    util_1.assertObject(x);
                    if (!x.value.has(prop)) {
                        x = value_1.NULL;
                        break;
                    }
                    else {
                        x = x.value.get(prop);
                    }
                }
                util_1.assertFunction(x);
                const args = await Promise.all(node.args.map(async (expr) => await this._eval(expr, scope)));
                return this._fn(x, args);
            }
            case 'index': {
                const arr = scope.get(node.arr);
                util_1.assertArray(arr);
                const i = await this._eval(node.i, scope);
                util_1.assertNumber(i);
                return arr.value[i.value - 1]; // TODO: 存在チェック
            }
            case 'fn': {
                return value_1.FN(node.args, node.children, scope);
            }
            case 'block': {
                return this._run(node.statements, scope.createChildScope());
            }
            case 'tmpl': {
                let str = '';
                for (const x of node.tmpl) {
                    if (typeof x === 'string') {
                        str += x;
                    }
                    else {
                        // TODO: Core:to_strと処理を共通化する
                        const v = await this._eval(x, scope);
                        let text = '';
                        if (v.type === 'str')
                            text = v.value;
                        else if (v.type === 'num')
                            text = v.value.toString();
                        str += text;
                    }
                }
                return value_1.STR(str);
            }
            case 'return': {
                const val = await this._eval(node.expr, scope);
                this.log('block:return', { scope: scope.name, val: val });
                return value_1.RETURN(val);
            }
            case 'ns': {
                return value_1.NULL; // nop
            }
            case 'meta': {
                return value_1.NULL; // nop
            }
        }
    }
    async _run(program, scope) {
        this.log('block:enter', { scope: scope.name });
        let v = value_1.NULL;
        for (let i = 0; i < program.length; i++) {
            const node = program[i];
            v = await this._eval(node, scope);
            if (v.type === 'return') {
                this.log('block:return', { scope: scope.name, val: v.value });
                return v;
            }
        }
        this.log('block:leave', { scope: scope.name, val: v });
        return v;
    }
    registerAbortHandler(handler) {
        this.abortHandlers.push(handler);
    }
    unregisterAbortHandler(handler) {
        this.abortHandlers = this.abortHandlers.filter(h => h != handler);
    }
    abort() {
        this.stop = true;
        for (const handler of this.abortHandlers) {
            handler();
        }
        this.abortHandlers = [];
    }
}
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "exec", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "execFn", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "log", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "collectNs", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "collectNsMember", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "_fn", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "_eval", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "_run", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "registerAbortHandler", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "unregisterAbortHandler", null);
__decorate([
    autobind_decorator_1.default
], AiScript.prototype, "abort", null);
__decorate([
    autobind_decorator_1.default
], AiScript, "collectMetadata", null);
exports.AiScript = AiScript;
//# sourceMappingURL=index.js.map